\ Copyright (c) 2005 by Offete Enterprises
\ All rights reserved 
\
\ (Translated from eForth 1.0 by Bill Muench and C. H. Ting, 1990)
\ GBAKERL.F   18MAY02 created by Wuja.
\ GBAeFORTH high-level definitions.
\ V1.00,   26OTC01WUJA,  Translate high-level defintions from eForth 1.0.
\ V2.00,   30JUN04CHT, Gerenal optimization with MOVE, and shift operations.
\ V5.02,   27sep04cht, GameGirl9 DSO host.
\ V5.04,   10nov04cht, GameGirlADUC interface. CONSOLE calls FILE.
\ V5.10,   19nov05cht, convert to ADuC metacompiler
\ V5.20,   22nov05cht, add startup.f, and get rid of Keil boot code

\ System and user variables

CODE SP0     ( -- a )              \ Pointer to bottom of the data stack.
  [-P!] TOS, SP, 4 #            ,STR        \ push TOS
  TOS, R3, $10 #                ,ADD        \ add user offset
  $NEXT 
CODE RP0     ( -- a )              \ Pointer to bottom of the return stack.
  [-P!] TOS, SP, 4 #            ,STR        \ push TOS
  TOS, R3, $14 #                ,ADD        \ add user offset
  $NEXT 
CODE '?KEY   ( -- a )              \ Execution vector of ?KEY.
  [-P!] TOS, SP, 4 #            ,STR        \ push TOS
  TOS, R3, $18 #                ,ADD        \ add user offset
  $NEXT 
CODE 'EMIT   ( -- a )              \ Execution vector of EMIT.
  [-P!] TOS, SP, 4 #            ,STR        \ push TOS
  TOS, R3, $1C #                ,ADD        \ add user offset
  $NEXT 
CODE 'EXPECT ( -- a )              \ Execution vector of EXPECT.
  [-P!] TOS, SP, 4 #            ,STR        \ push TOS
  TOS, R3, $20 #                ,ADD        \ add user offset
  $NEXT 
CODE 'TAP    ( -- a )              \ Execution vector of TAP.
  [-P!] TOS, SP, 4 #            ,STR        \ push TOS
  TOS, R3, $24 #                ,ADD        \ add user offset
  $NEXT 
CODE 'ECHO   ( -- a )              \ Execution vector of ECHO.
  [-P!] TOS, SP, 4 #            ,STR        \ push TOS
  TOS, R3, $28 #                ,ADD        \ add user offset
  $NEXT 
CODE 'PROMPT ( -- a )              \ Execution vector of PROMPT.
  [-P!] TOS, SP, 4 #            ,STR        \ push TOS
  TOS, R3, $2C #                ,ADD        \ add user offset
  $NEXT 
CODE BASE    ( -- a )              \ Storage of the radix base for numeric I/O.
  [-P!] TOS, SP, 4 #            ,STR        \ push TOS
  TOS, R3, $30 #                ,ADD        \ add user offset
  $NEXT 
CODE tmp     ( -- a )              \ A temporary storage location used in parse and find.
  [-P!] TOS, SP, 4 #            ,STR        \ push TOS
  TOS, R3, $34 #                ,ADD        \ add user offset
  $NEXT COMPILE-ONLY
CODE SPAN    ( -- a )              \ Hold character count received by EXPECT.
  [-P!] TOS, SP, 4 #            ,STR        \ push TOS
  TOS, R3, $38 #                ,ADD        \ add user offset
  $NEXT 
CODE >IN     ( -- a )              \ Hold the character pointer while parsing input
  [-P!] TOS, SP, 4 #            ,STR        \ push TOS
  TOS, R3, $3C #                ,ADD        \ add user offset
  $NEXT 
CODE #TIB    ( -- a )              \ Hold the current count. 
  [-P!] TOS, SP, 4 #            ,STR        \ push TOS
  TOS, R3, $40 #                ,ADD        \ add user offset
  $NEXT 
CODE CSP     ( -- a )              \ Hold the stack pointer for error checking.
  [-P!] TOS, SP, 4 #            ,STR        \ push TOS
  TOS, R3, $48 #                ,ADD        \ add user offset
  $NEXT 
CODE 'EVAL   ( -- a )              \ Execution vector of EVAL.
  [-P!] TOS, SP, 4 #            ,STR        \ push TOS
  TOS, R3, $4C #                ,ADD        \ add user offset
  $NEXT 
CODE 'NUMBER ( -- a )              \ Execution vector of NUMBER?.
  [-P!] TOS, SP, 4 #            ,STR        \ push TOS
  TOS, R3, $50 #                ,ADD        \ add user offset
  $NEXT 
CODE HLD     ( -- a )              \ Hold a pointer in building a numeric output string.
  [-P!] TOS, SP, 4 #            ,STR        \ push TOS
  TOS, R3, $54 #                ,ADD        \ add user offset
  $NEXT 
CODE HANDLER ( -- a )              \ Hold the return stack pointer for error handling.
  [-P!] TOS, SP, 4 #            ,STR        \ push TOS
  TOS, R3, $58 #                ,ADD        \ add user offset
  $NEXT 
CODE CONTEXT ( -- a )              \ A area to specify vocabulary search order.
  [-P!] TOS, SP, 4 #            ,STR        \ push TOS
  TOS, R3, $5C #                ,ADD        \ add user offset
  $NEXT 
CODE CURRENT ( -- a )              \ Point to the vocabulary to be extended.
  [-P!] TOS, SP, 4 #            ,STR        \ push TOS
  TOS, R3, $80 #                ,ADD        \ add user offset
  $NEXT 
CODE CP      ( -- a )              \ Point to the top of the code dictionary.
  [-P!] TOS, SP, 4 #            ,STR        \ push TOS
  TOS, R3, $88 #                ,ADD        \ add user offset
  $NEXT 
CODE NP      ( -- a )              \ Point to the bottom of the name dictionary.
  [-P!] TOS, SP, 4 #            ,STR        \ push TOS
  TOS, R3, $8C #                ,ADD        \ add user offset
  $NEXT 
CODE LAST    ( -- a )              \ Point to the last name in the name dictionary.
  [-P!] TOS, SP, 4 #            ,STR        \ push TOS
  TOS, R3, $90 #                ,ADD        \ add user offset
  $NEXT 
CODE 'FORTH    ( -- a )              \ Point to the last name in the name dictionary.
  [-P!] TOS, SP, 4 #            ,STR        \ push TOS
  TOS, R3, $94 #                ,ADD        \ add user offset
  $NEXT 

\ Common functions

: FORTH 'FORTH CONTEXT ! ;

: WITHIN ( u ul uh -- t )                   \ Return true if u is within the range of ul and uh.
  OVER - >R                                 \ ul <= u < uh
  - R> U< ;      

CODE UM/MOD ( udl udh u -- ur uq )          \ Unsigned divide of a double by a single. Return mod and quotient.
  YP, 1 #                       ,MOV        \ YP := 1  
  [] WP, SP, 4 #                ,LDR        \ WP := [SP](udh) , SP := SP+4
  [P] XP, SP                     ,LDR        \ XP := [SP](udl)  
  
  S, XP, XP, XP                 ,ADD        \ XP := XP*2
  S, WP, WP, WP                 ,ADC        \ WP := WP*2+C
  CC, PC, PC, 8 #               ,ADD        \ jump to normal sub  
  WP, WP, TOS                   ,SUB        \ WP := WP-TOS (if carry)
  XP, XP, 1 #                   ,ADD        \ XP := XP+1
  PC, PC, C #                   ,ADD        \ jump to check
  S, WP, WP, TOS                ,SUB        \ WP := WP-TOS
  CS, XP, XP, 1 #               ,ADD        \ XP := XP+1   
  CS, PC, PC, 0 #               ,ADD        \ jump to check 
  WP, WP, TOS                   ,ADD        \ restore WP 
  S, YP, YP, YP                 ,ADD        \ YP := YP*2
  CC, PC, PC, 34 #              ,SUB        \ go back loop
  
  TOS, XP                       ,MOV        \ TOS := XP
  [P] WP, SP                     ,STR        \ [SP] := WP
  $NEXT
END-CODE   

: M/MOD ( d n -- r q )                      \ Signed floored divide of double by single. Return mod and quotient.
  DUP 0<  DUP >R
    IF NEGATE >R DNEGATE R>
    THEN >R DUP 0< IF R@ + THEN R> UM/MOD R>
    IF SWAP NEGATE SWAP 
    THEN 
  ;

: /MOD ( n n -- r q )                       \ Signed divide. Return mod and quotient.
  OVER 0< SWAP M/MOD ;

: MOD ( n n -- r )                          \ Signed divide. Return mod only.
  /MOD DROP ;

: / ( n n -- q )                            \ Signed divide. Return quotient only.
  /MOD SWAP DROP ;

: */MOD ( n1 n2 n3 -- r q )                 \ Multiply n1 and n2, then divide by n3. Return mod and quotient.
  >R M* R> M/MOD ;

: */ ( n1 n2 n3 -- q )                      \ Multiply n1 by n2, then divide by n3. Return quotient only.
  */MOD SWAP DROP ;

\ Miscellaneous

CODE ALIGNED ( b -- a )                        \ Align address to the cell boundary.
  TOS, TOS, 3 # ,ADD
  WP, 3 # ,MVN
  TOS, TOS, WP ,AND
  $NEXT

: >CHAR ( c -- c )                          \ Filter non-printing characters.
[  [-P!] TOS, SP, 4 #            ,STR        \ push TOS
  TOS, $7F #                    ,MOV
]  AND DUP BL
[  [-P!] TOS, SP, 4 #            ,STR        \ push TOS
  TOS, $7F #                    ,MOV
]  WITHIN NOT
  IF DROP 
[  [-P!] TOS, SP, 4 #            ,STR        \ push TOS
  TOS, $5F #                    ,MOV
]  THEN
  ;

: DEPTH ( -- n )                            \ Return the depth of the data stack.
  SP@ SP0 @ SWAP - CELL/ ;

: PICK ( ... +n -- ... w )                  \ Copy the nth stack item to tos.
  1+ CELLS SP@ + @ ;

\ Memory access

: HERE ( -- a )                             \ Return the top of the code dictionary.
  CP @ ;

: PAD ( -- a )                              \ Return the address of a temporary buffer.
  HERE 
[  TOS, TOS, $50 # ,ADD   ]
 ;

: TIB ( -- a )                              \ Return the address of the terminal input buffer.
  #TIB CELL+ @ ;

CODE @EXECUTE ( a -- )                         \ Execute vector stored in address a.
  S, WP, TOS  			,MOV
  [] TOS, SP, 4 #               ,LDR        \ pop TOS
  [P] WP, WP 			,LDR
  NE, PC, WP 			,MOV
  $NEXT

CODE CMOVE ( b1 b2 u -- )                   \ Copy u bytes from b1 to b2.
  [] XP, SP, 4 #                ,LDR        \ XP := [SP](b2)  , SP := SP+4            
  [] YP, SP, 4 #                ,LDR        \ YP := [SP](b1)  , SP := SP+4
  PC, PC, 4 #                   ,ADD        \ skip first time  
  [B] WP, YP, 1 #               ,LDR        \ WP := byte[YP]  , YP := YP+1
  [B] WP, XP, 1 #               ,STR        \ [XP] := byte WP , XP := XP+1
  S, TOS, TOS                   ,MOV        \ TOS := TOS
  EQ, PC, PC, 4 #               ,ADD        \ PC := PC+4
  TOS, TOS, 1 #                 ,SUB        \ TOS := TOS-1
  PC, PC, 1C #                  ,SUB        \ go back loop
  [] TOS, SP, 4 #               ,LDR        \ TOS := [SP]     , SP := SP+4
  $NEXT
END-CODE  

CODE MOVE ( a1 a2 u -- )                   \ Copy u bytes from b1 to b2.
  [-] TOS, TOS, 4 #		,AND
  [] XP, SP, 4 #                ,LDR        \ XP := [SP](b2)  , SP := SP+4            
  [] YP, SP, 4 #                ,LDR        \ YP := [SP](b1)  , SP := SP+4
  PC, PC, 4 #                   ,ADD        \ skip first time  
  [] WP, YP, 4 #                ,LDR        \ WP := byte[YP]  , YP := YP+1
  [] WP, XP, 4 #                ,STR        \ [XP] := byte WP , XP := XP+1
  S, TOS, TOS                   ,MOV        \ TOS := TOS
  EQ, PC, PC, 4 #               ,ADD        \ PC := PC+4
  TOS, TOS, 4 #                 ,SUB        \ TOS := TOS-1
  PC, PC, 1C #                  ,SUB        \ go back loop
  [] TOS, SP, 4 #               ,LDR        \ TOS := [SP]     , SP := SP+4
  $NEXT
END-CODE  

CODE FILL ( b u c -- )                         \ Fill u bytes of character c to area beginning at b.
  [] XP, SP, 4 #                ,LDR        \ XP := [SP](b2)  , SP := SP+4            
  [] YP, SP, 4 #                ,LDR        \ YP := [SP](b1)  , SP := SP+4
  PC, PC, 4 #                   ,ADD        \ skip first time  
  TOS, TOS               	,MOV        \ WP := byte[YP]  , YP := YP+1
  [B] TOS, YP, 1 #              ,STRB        \ [XP] := byte WP , XP := XP+1
  S, XP, XP                   	,MOV        \ TOS := TOS
  EQ, PC, PC, 4 #               ,ADD        \ PC := PC+4
  XP, XP, 1 #                 	,SUB        \ TOS := TOS-1
  PC, PC, 1C #                  ,SUB        \ go back loop
  [] TOS, SP, 4 #               ,LDR        \ TOS := [SP]     , SP := SP+4
  $NEXT
END-CODE  

: -TRAILING ( b u -- b u )                  \ Adjust the count to eliminate trailing white space.
  FOR NOP
    ( AFT ) BL OVER R@ + C@ <
        IF R> 1+ EXIT THEN 
    ( THEN )
  NEXT 
[  [-P!] TOS, SP, 4 #            ,STR        \ push TOS
  TOS 0 # ,MOV
]  ;

: PACK$ ( b u a -- a )                      \ Build a counted string with u characters from b. Null fill.
  ALIGNED DUP >R
  OVER + 1+
[  [-P!] TOS, SP, 4 #            ,STR        \ push TOS
  TOS 3 # ,MVN
]  AND
[  [-P!] TOS, SP, 4 #            ,STR        \ push TOS
  TOS 0 # ,MOV
]  SWAP ! R@
  2DUP C! 1+
  SWAP CMOVE R> ;

\ Numeric output, single precision
: DIGIT ( u -- c )                          \ Convert digit u to a character.
[  [-P!] TOS, SP, 4 #            ,STR        \ push TOS
  TOS 9 # ,MOV
]  OVER < 
[  TOS, TOS 7 # ,AND
]  + 
[  TOS, TOS $30 # ,ADD
]  ;

: EXTRACT ( n base -- n c )                 \ Extract the least significant digit from n.
[  [-P!] TOS, SP, 4 #            ,STR        \ push TOS
  TOS 0 # ,MOV
]  SWAP UM/MOD SWAP DIGIT ;

: <# ( -- )                                 \ Initiate the numeric output process.
  PAD HLD ! ;

: HOLD ( c -- )                             \ Insert a character into the numeric output string.
  HLD @ 1- DUP HLD ! C! ;

: # ( u -- u )                              \ Extract one digit from u and append the digit to output string.
  BASE @ EXTRACT HOLD ;              

: #S ( u -- 0 )                             \ Convert u until all digits are added to the output string.
  BEGIN # DUP WHILE 
  [ bra THEN ] ;

: SIGN ( n -- )                             \ Add a minus sign to the numeric output string.
  0< IF 
[  [-P!] TOS, SP, 4 #_            ,STR        \ push TOS
  TOS $2D #_ ,MOV
]  HOLD THEN ;                

: #> ( w -- b u )                           \ Prepare the output string to be TYPE'd.
  DROP HLD @ PAD OVER - ;

: str ( n -- b u )                          \ Convert a signed integer to a numeric string.
  DUP >R ABS <# #S R> SIGN #> ;

: HEX ( -- )                                \ Use radix 16 as base for numeric conversions.
[  [-P!] TOS, SP, 4 #_            ,STR        \ push TOS
  TOS $10 #_ ,MOV
]  BASE ! ;

: DECIMAL ( -- )                            \ Use radix 10 as base for numeric conversions.
[  [-P!] TOS, SP, 4 #_            ,STR        \ push TOS
  TOS $0A #_ ,MOV
]  BASE ! ;

\ Numeric input, single precision
: DIGIT? ( c base -- u t )                  \ Convert a character to its numeric value. A flag indicates success.
  >R 
[  [-P!] TOS, SP, 4 #_            ,STR        \ push TOS
  TOS $30 #_ ,MOV
]  - 
[  [-P!] TOS, SP, 4 #_            ,STR        \ push TOS
  TOS $9 #_ ,MOV
]  OVER <
  IF 
[  [-P!] TOS, SP, 4 #_            ,STR        \ push TOS
  TOS $7 #_ ,MOV
]  - DUP 
[  [-P!] TOS, SP, 4 #_            ,STR        \ push TOS
  TOS $0A #_ ,MOV
]  < OR 
  THEN 
  DUP R> U< ;

: NUMBER? ( a -- n T | a F )                \ Convert a number string to integer. Push a flag on tos.
  BASE @ >R  0 OVER COUNT ( a 0 b n)
  OVER C@ [ CHAR $ ] LITERAL =
  IF HEX SWAP 1+ SWAP 1- THEN ( a 0 b' n')
  OVER C@ [ CHAR - ] LITERAL = >R ( a 0 b n)
  SWAP R@ - SWAP R@ + ( a 0 b" n") ?DUP
  IF 1- ( a 0 b n)
      FOR DUP >R C@ BASE @ DIGIT?
        WHILE SWAP BASE @ * +  R> 1+
      NEXT R@ ( ?sign) SWAP DROP ( b) 
        IF NEGATE THEN SWAP
        ELSE R> R> ( b index) 2DROP ( digit number) 2DROP 0
        THEN DUP
  THEN
  R> ( n ?sign) 2DROP R> BASE ! ;

\ Basic I/O
: ?KEY ( -- c T | F )                       \ Return input character and true, or a false if no input.
  '?KEY @EXECUTE ;                

: KEY ( -- c )                              \ Wait for and return an input character.
  BEGIN ?KEY UNTIL ;

: EMIT ( c -- )                             \ Send a character to the output device.
  'EMIT @EXECUTE ;

: NUF? ( -- t )                             \ Return false if no input, else pause and if CR return true.
  ?KEY DUP 
  IF 2DROP KEY 0D = 
  THEN
  ;

: PACE ( -- )                               \ Send a pace character for the file downloading process.
  0B EMIT ;

: SPACE ( -- )                              \ Send the blank character to the output device.
  BL EMIT ;

: SPACES ( +n -- )                          \ Send n spaces to the output device.
  0 MAX FOR 
      AFT SPACE THEN 
  NEXT ;

: TYPE ( b u -- )                           \ Output u characters from b.
  FOR 
    AFT COUNT EMIT THEN 
  NEXT DROP ;

: CR ( -- )                                 \ Output a carriage return and a line feed.
  0D EMIT 0A EMIT ; 

: do$ ( -- a )                              \ Return the address of a compiled string.
  R> R@ R> COUNT + ALIGNED >R SWAP >R ; COMPILE-ONLY

: $"| ( -- a )                              \ Run time routine compiled by $". Return address of a compiled string.
  do$ ; COMPILE-ONLY

: .$ ( a -- ) COUNT TYPE ;

: ."| ( -- )                                \ Run time routine of ." . Output a compiled string.
  do$ .$ ; COMPILE-ONLY

: .R ( n +n -- )                            \ Display an integer in a field of n columns, right justified.
  >R str R> OVER - SPACES TYPE ;

: U.R ( u +n -- )                           \ Display an unsigned integer in n column, right justified.
  >R <# #S #> R> OVER - SPACES TYPE ;

: U. ( u -- )                               \ Display an unsigned integer in free format.
  <# #S #> SPACE TYPE ;

: . ( w -- )                                \ Display an integer in free format, preceeded by a space.
  BASE @ 0A XOR 
    IF U. EXIT 
    THEN 
  str SPACE TYPE ;

: ? ( a -- )                                \ Display the contents in a memory cell.
  @ . ;

\ Parsing
: parse ( b u c -- b u delta ; <string> ) \ Scan string delimited by c. Return found string and its offset.
  tmp !  OVER >R  DUP
  IF 1-  tmp @ BL =
    IF 
      FOR BL OVER C@ - 0< NOT
        WHILE 1+
      NEXT R> DROP 0 DUP EXIT
        THEN  R>
    THEN OVER SWAP \ b' b' u' \ 'scan'
    FOR tmp @ OVER C@ - tmp @ BL =
      IF 0< THEN WHILE 1+
    NEXT DUP >R  
      ELSE R> DROP DUP 1+ >R
      THEN OVER -  R>  R> - EXIT
  THEN ( b u) OVER R> - ;

: PARSE ( c -- b u ; <string> )             \ Scan input stream and return counted string delimited by c.
  >R  TIB >IN @ +  #TIB @ >IN @ -  R> parse >IN +! ;

ONLY META ALSO TARGET ALSO hidden ALSO DEFINITIONS

: .( 
  [ CHAR ) ] LITERAL PARSE TYPE ; IMMEDIATE

: ( 
  [ CHAR ) ] LITERAL PARSE 2DROP ; IMMEDIATE

: \ 
  $13 PARSE 2DROP ; IMMEDIATE

: CHAR
  BL PARSE DROP C@ ;

ONLY META ALSO TARGET ALSO DEFINITIONS

: WORD ( c -- a ; <string> )                \ Parse a word from input stream and copy it to code dictionary.
  PARSE HERE CELL+ PACK$ ;

: TOKEN ( -- a ; <string> )                 \ Parse a word from input stream and copy it to name dictionary.
  BL WORD ;
  
\ Dictionary search
: NAME> ( na -- ca )                        \ Return a code address given a name address.
  COUNT 1F AND + ALIGNED ;

: SAME? ( a a u -- a a f \ -0+ )            \ Compare u cells in two strings. Return 0 if identical.
  FOR AFT OVER R@ CELLS + @
          OVER R@ CELLS + @ -  ?DUP
    IF R> DROP EXIT THEN 
  THEN NEXT 0 ;

: find ( a va -- ca na | a F )              \ Search a vocabulary for a string. Return ca and na if succeeded.
  SWAP              \ va a
  DUP C@ 4 / tmp ! \ va a  \ get cell count
  DUP @ >R          \ va a  \ count byte & 1st char
  CELL+ SWAP        \ a' va
  BEGIN @ DUP       \ a' na na
    IF DUP @ [ =MASK ] LITERAL AND  R@ XOR \ ignore lexicon bits
      IF CELL+ -1 ELSE CELL+ tmp @ SAME? THEN
    ELSE R> DROP SWAP CELL- SWAP EXIT
    THEN
  WHILE 2 CELLS -   \ a' la
  REPEAT R> DROP SWAP DROP CELL- DUP NAME> SWAP ;

: NAME? ( a -- ca na | a F )                \ Search all context vocabularies for a string.
  CONTEXT  DUP 2@ XOR IF CELL- THEN >R  \ context<>also
  BEGIN R>  CELL+  DUP >R  @  ?DUP
  WHILE find  ?DUP
  UNTIL R> DROP EXIT THEN R> DROP  0 ;

\ Terminal response
: ^H ( bot eot cur -- bot eot cur )         \ Backup the cursor by one character.
  >R OVER R> SWAP OVER XOR
  IF 08 'ECHO @EXECUTE 1- BL 'ECHO @EXECUTE  \ back, fill blank, and back again
     08 'ECHO @EXECUTE 
  THEN
  ;

: TAP ( bot eot cur c -- bot eot cur )      \ Accept and echo the key stroke and bump the cursor.
  DUP 'ECHO @EXECUTE OVER C! 1+ ;

: kTAP ( bot eot cur c -- bot eot cur )     \ Process a key stroke, CR or backspace.
  DUP 0D XOR
    IF 08 XOR 
      IF BL TAP 
      ELSE ^H 
      THEN EXIT
    THEN 
  DROP SWAP DROP DUP ;

: accept ( b u -- b u )                     \ Accept characters to input buffer. Return with actual count.
  OVER + OVER
  BEGIN 2DUP XOR
  WHILE KEY DUP BL 7F WITHIN  
    IF TAP ELSE 'TAP @EXECUTE THEN
  REPEAT DROP OVER - ;

: EXPECT ( b u -- )                         \ Accept input stream and store count in SPAN.
  'EXPECT @EXECUTE SPAN ! DROP ;

: QUERY  ( -- )                             \ Accept input stream to terminal input buffer.
  TIB 50 'EXPECT @EXECUTE #TIB !  DROP 0 >IN ! ;

\ Error handling
: CATCH ( ca -- 0 | err# )                  \ Execute word at ca and set up an error frame for it.
  SP@ >R  HANDLER @ >R  RP@ HANDLER !
  EXECUTE
  R> HANDLER !  R> DROP  0 ;

: THROW ( err# -- err# )                    \ Reset system to current local error frame an update error flag.
  HANDLER @ RP!  R> HANDLER !  R> SWAP >R SP! DROP R> ;

VARIABLE NULL$   \ 0 NULL$ !

: ABORT ( -- )                              \ Reset data stack and jump to QUIT.
  NULL$ THROW ;

: abort"| ( f -- )                          \ Run time routine of ABORT" . Abort with a message.
  IF do$ THROW 
  THEN do$ DROP ; COMPILE-ONLY

[FORTH] : ABORT" 	[TARGET] [COMPILE] abort"| $LIT [FORTH] ; IMMEDIATE [TARGET]
[FORTH] : $" 	[TARGET] [COMPILE]     $"| $LIT [FORTH] ; IMMEDIATE [TARGET]
[FORTH] : ." 	[TARGET] [COMPILE]     ."| $LIT [FORTH] ; IMMEDIATE [TARGET]

\ The text interpreter
: $INTERPRET  ( a -- )                      \ Interpret a word. If failed, try to convert it to an integer.
  NAME? ?DUP
  IF @ [ =COMPILE-ONLY ] LITERAL AND
    ABORT" compile ONLY" EXECUTE EXIT
  THEN
  'NUMBER @EXECUTE
  IF EXIT THEN THROW ;

ONLY META ALSO TARGET ALSO hidden DEFINITIONS
: [ [META]
  [ ' $INTERPRET ] LITERAL 'EVAL ! ; IMMEDIATE
ONLY META ALSO TARGET ALSO DEFINITIONS

: .OK ( -- )                                \ Display 'ok' only while interpreting.
  [ ' $INTERPRET ] LITERAL 'EVAL @ = 
    IF ."  ok" 
    THEN 
  CR ;

: ?STACK ( -- )                             \ Abort if the data stack underflows.
  DEPTH 0< ABORT" underflow" ;

: EVAL ( -- )                               \ Interpret the input stream.
  BEGIN TOKEN DUP C@
  WHILE 'EVAL @EXECUTE ?STACK
  REPEAT DROP 'PROMPT @EXECUTE ;

\ Shell
: PRESET ( -- )                             \ Reset data stack pointer and the terminal input buffer.
  SP0 @ SP!  [ =TIB ] LITERAL #TIB CELL+ ! ;

: XIO ( a a a -- )                          \ Reset the I/O vectors 'EXPECT, 'TAP, 'ECHO and 'PROMPT.
  [ ' accept ] LITERAL 'EXPECT !
  'TAP !  'ECHO !  'PROMPT ! ;

: FILE ( -- )                               \ Select I/O vectors for file download.
  [ ' PACE ] LITERAL [ ' DROP ] LITERAL [ ' kTAP ] LITERAL XIO ;

: HAND ( -- )                               \ Select I/O vectors for terminal interface.
  [ ' .OK  ] LITERAL [ ' TX!  ] LITERAL [ ' kTAP ] LITERAL XIO ;

CREATE I/O  ' ?RX ,   ' TX! ,              \ defaults

: CONSOLE ( -- )                            \ Initiate terminal interface.
  I/O 2@ '?KEY 2! HAND ;		    \ for GBA-DSO interface

: QUIT ( -- )                               \ Reset return stack pointer and start text interpreter.
  PRESET CONSOLE                            \ Added for READ and OK
  RP0 @ RP!
  BEGIN ( [COMPILE] ) [hidden] [ [TARGET]
    BEGIN QUERY [ ' EVAL ] LITERAL CATCH ?DUP
    UNTIL ( a)
    'PROMPT @ SWAP
    CONSOLE  NULL$ OVER XOR
    IF SPACE COUNT TYPE ." ?" THEN
    [ ' .OK ] LITERAL XOR
    IF 1B EMIT THEN
    PRESET
  AGAIN ;

: OVERT ( -- )                              \ Link a new word into the current vocabulary.
  LAST @ CURRENT @  ! ;

\ Tools
: _TYPE ( b u -- )                          \ Display a string. Filter non-printing characters.
  FOR 
    AFT COUNT >CHAR EMIT THEN 
  NEXT DROP ;

: dm+ ( a u -- a )                          \ Dump u bytes from a, leaving a+u on the stack.
  OVER 4 U.R SPACE CELL/
  FOR AFT 
    DUP @ 9 U.R 
    CELL+
  THEN NEXT ;
 
: DUMP ( a u -- )                           \ Dump u bytes from a, in a formatted manner.
  BASE @ >R HEX  10 /
  FOR CR 10 2DUP dm+ ROT ROT 2 SPACES _TYPE NUF? 0 = WHILE
  NEXT ELSE R> DROP THEN DROP  R> BASE ! ;

: .S ( ... -- ... )                         \ Display the contents of the data stack.
  CR DEPTH FOR AFT R@ PICK . THEN NEXT ."  <tos" ;

: !CSP ( -- )                               \ Save stack pointer in CSP for error checking.
  SP@ CSP ! ;

: ?CSP ( -- )                               \ Abort if stack pointer differs from that saved in CSP.
  SP@ CSP @ XOR ABORT" stack depth" ;

ONLY META ALSO TARGET ALSO hidden DEFINITIONS
: '
  TOKEN NAME? IF EXIT THEN THROW ;
ONLY META ALSO TARGET ALSO DEFINITIONS

: >NAME ( ca -- na | F )                    \ Convert code address to a name address.
  CURRENT
    BEGIN CELL+ @ ?DUP WHILE 2DUP
      BEGIN @ DUP WHILE 2DUP NAME> XOR
      WHILE 1 CELLS -
      REPEAT      THEN SWAP DROP ?DUP
    UNTIL 
  SWAP DROP SWAP DROP EXIT THEN 0 SWAP DROP ;

: .ID ( na -- )                             \ Display the name at address.
  ?DUP IF COUNT 001F AND TYPE EXIT THEN ." {noName}" ;

: SEE ( -- ; <string> )                     \ A simple decompiler.
  [hidden] ' [TARGET] CR 
  BEGIN
    20 FOR
    CELL+ DUP @ DUP
    800000 AND
    IF FF000000 OR ELSE FFFFFF AND THEN
    2+ CELLS OVER + >NAME ?DUP
    IF SPACE .ID
    ELSE DUP @ U.
    THEN
    NEXT
    KEY 0A < 
  UNTIL
  DROP ;

: WORDS ( -- )                              \ Display the names in the context vocabulary.
  CR  CONTEXT @
  BEGIN @ ?DUP
  WHILE DUP SPACE .ID 1 CELLS -  NUF?
  UNTIL DROP THEN ;

: READ ( -- )
  PAD 100 +
  BEGIN 
    KEY 
     DUP 5C =
          IF DROP 20 BEGIN KEY 0D = UNTIL THEN
     DUP 0D =
          IF DROP 20 THEN
     DUP 0A =
          IF DROP 20 THEN
     DUP 1A XOR
    WHILE OVER C! 1+
  REPEAT DROP 
  PAD 100 + - SPAN ! ;
  
: OK ( -- )
  PAD 100 + #TIB CELL+ ! SPAN @ #TIB ! 0 >IN !
  EVAL QUIT ;

\ Hardware reset
CREATE VER  506 ,

: hi ( -- )                                 \ Display the sign-on message of eForth.
   CR ." ADuC eForth v5.20"
   CR ." Copyright (c) 2005 by Offete Enterprises"
   CR ." All Right Reserved"
   CR 
  ; COMPILE-ONLY

CREATE 'BOOT   ' hi ,

[META]
CODE INIT_REG ( sp -- )
  R2, TOS      ,MOV \ RP=11E00
  R3, R2 200 # ,SUB \ UP=11C00
  R1, R2 200 # ,SUB \ SP=11C00
  $NEXT
[TARGET]

: COLD ( -- )                               \ The hilevel cold start sequence.
  11E00 INIT_REG
  !IO \ initialize IO device
  20 FFFF0500 ! \ ADCCON
  1 FFFF0480 ! \ REFCON
  11 FFFFF404 ! \ GP1CON
  10100000 FFFFF400 ! \ GP0CON
  80 FFFF0300 ! \ T0LD
  0 FFFF0308 ! \ T0CON
  4000000 FFFFF460 ! \ GP4DAT
  80200 11C00 A0 CMOVE \ USER AREA
  HEX
  BEGIN 
    FORTH CONTEXT @ DUP
    CURRENT 2! OVERT
    PRESET
    'BOOT @EXECUTE
\    WORDS CR
\    11C00 A0 DUMP
    QUIT
  AGAIN ;

